/*
 * eth_ecat_pdo_fmmu.hpp
 *
 *  Created on: May 16, 2023
 *      Author: algin
 */

#ifndef FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_PDO_FMMU_HPP_
#define FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_PDO_FMMU_HPP_

#include "semaphore/semaphore.hpp"
#include "mutex/mutex.hpp"

#include "ethernet_industry/eth_ecat_buffer.hpp"
#include "ethernet_industry/eth_ecat_custom_promise.hpp"

namespace free_rtos {

namespace ecat_pdo_fmmu {

class EcatPdoFMMUSlave {
public:
    EcatPdoFMMUSlave(ecat_buffer::EcatBufferSlave& buffer_slave)
    : buffer_slave_(buffer_slave) { }

    template<typename DataType>
    datagram::EcatDatagram<command::LWR, DataType> make_datagram_write(DataType& data) {
        using TDatagram = datagram::EcatDatagram<command::LWR, DataType>;

        ecat_buffer::FMMUProperties& fmmu_properties = buffer_slave_.get_fmmu_properties_write();
        address::Logical logical = fmmu_properties.address;

        return TDatagram{ {{logical}}, data };
    }

    template<typename DataType>
    datagram::EcatDatagram<command::LRD, DataType> make_datagram_read(DataType& data) {
        using TDatagram = datagram::EcatDatagram<command::LRD, DataType>;

        ecat_buffer::FMMUProperties& fmmu_properties = buffer_slave_.get_fmmu_properties_read();
        address::Logical logical = fmmu_properties.address;

        return TDatagram{ {{logical}}, data };
    }

private:
    ecat_buffer::EcatBufferSlave& buffer_slave_;
};

/*
// Функтор для обхода и передачи датаграмм в custom_tuple
struct DatagramFunctor {
    telegram::EcatTelegram& telegram;
    size_t number_of_slaves;
    datagram::IEcatDatagram *next;

    void operator()(datagram::IEcatDatagram& datagram) {
        if(next != nullptr) {
           datagram + *next;
        }

        next = &datagram;
    }

    void operator()() {
        if(next == nullptr) {
            return;
        }

        do {
            telegram.transfer(*next);
        } while(next->get_all_wkc() < number_of_slaves);
    }
};
*/
class EthEcatPdoFMMU {
public:
    EthEcatPdoFMMU(ecat_buffer::EthEcatBuffer& ecat_buffer): ecat_buffer_{ecat_buffer} { }

    void init();
    void process();

    template<typename... DataTypes>
    void pdo_write(address::Offset offset, DataTypes&... data) {
        custom_promise::WritePromise<DataTypes...> promise{offset, data...};

        mutex_write_.lock();
        //last_write_ = (*last_write_) + &promise;
        queue_write_ + promise;
        mutex_write_.unlock();

        promise.get_future().get();
    }

    template<typename... DataTypes>
    void pdo_read(address::Offset offset, DataTypes&... data) {
        custom_promise::ReadPromise<DataTypes...> promise{offset, data...};

        mutex_read_.lock();
        //last_read_ = (*last_read_) + &promise;
        queue_read_ + promise;
        mutex_read_.unlock();

        promise.get_future().get();
    }


    void pdo_write_async(custom_promise::IPromise& promise) {
        mutex_write_.lock();

        //last_write_ = (*last_write_) + &promise;
        queue_write_ + promise;

        mutex_write_.unlock();
    }


    void pdo_read_async(custom_promise::IPromise& promise) {
        mutex_read_.lock();

        //last_read_ = (*last_read_) + &promise;
        queue_read_ + promise;

        mutex_read_.unlock();
    }

private:
    ecat_buffer::EthEcatBuffer& ecat_buffer_;

    std::vector<EcatPdoFMMUSlave> pdo_fmmu_slaves_;

    Mutex mutex_write_;
    Mutex mutex_read_;

    custom_promise::WritePromise<> queue_write_{0};
    custom_promise::ReadPromise<> queue_read_{0};

    //custom_promise::IPromise *last_write_{&queue_write_};
    //custom_promise::IPromise *last_read_{&queue_read_};

    void wait_op();

    void process_write_queue(uint8_t* process_data, uint32_t len);
    void process_read_queue(uint8_t* process_data, uint32_t len);

    template<typename... DataTypes>
    void write(address::Offset offset, DataTypes&... data) {
        using TDatagram = datagram::EcatDatagram<command::LWR, DataTypes...>;

        telegram::EcatTelegram& telegram = ecat_buffer_.get_ecat().get_telegram();
        ecat_buffer::FMMUGlobalProperties& fmmu_global_properties = ecat_buffer_.get_fmmu_global_properties();
        address::Logical logical = fmmu_global_properties.logical_start_address + offset;
        TDatagram datagram{ {{logical}}, data... };

        do {
            telegram.transfer(datagram);
        } while(datagram.get_wkc() < 0x0001);
    }

    template<typename... DataTypes>
    void read(address::Offset offset, DataTypes&... data) {
        using TDatagram = datagram::EcatDatagram<command::LRD, DataTypes...>;

        telegram::EcatTelegram& telegram = ecat_buffer_.get_ecat().get_telegram();
        ecat_buffer::FMMUGlobalProperties& fmmu_global_properties = ecat_buffer_.get_fmmu_global_properties();
        address::Logical logical = fmmu_global_properties.logical_start_address + fmmu_global_properties.logical_full_length_write + offset;
        TDatagram datagram{ {{logical}}, data... };

        do {
            telegram.transfer(datagram);
        } while(datagram.get_wkc() < 0x0001);
    }

    template<typename... DataTypes>
    void read_write(address::Offset offset_read, address::Offset offset_write, DataTypes&... data) {
        using TDatagramRead = datagram::EcatDatagram<command::LRD, DataTypes...>;
        using TDatagramWrite = datagram::EcatDatagram<command::LWR, DataTypes...>;

        telegram::EcatTelegram& telegram = ecat_buffer_.get_ecat().get_telegram();
        ecat_buffer::FMMUGlobalProperties& fmmu_global_properties = ecat_buffer_.get_fmmu_global_properties();

        address::Logical logical_read = fmmu_global_properties.logical_start_address + fmmu_global_properties.logical_full_length_write + offset_read;
        TDatagramRead datagram_read{ {{logical_read}}, data... };

        address::Logical logical_write = fmmu_global_properties.logical_start_address + offset_write;
        TDatagramWrite datagram_write{ {{logical_write}}, data... };

        datagram_read + datagram_write;

        do {
            telegram.transfer(datagram_read);
        } while(datagram_read.get_all_wkc() < 0x0001);
    }

/*
    // Запись PDO замапленных каждым слейвом
    template<typename... DataTypes>
    void write(DataTypes&... data) {
        if(sizeof...(data) > pdo_fmmu_slaves_.size()) {
            return;
        }

        size_t i = 0;
        custom_tuple<datagram::EcatDatagram<command::LWR, DataTypes>...> datagram_tuple{pdo_fmmu_slaves_[i++].make_datagram_write(data)...};
        DatagramFunctor functor{ecat_buffer_.get_ecat().get_telegram(), pdo_fmmu_slaves_.size(), nullptr};

        for_each_reverse(datagram_tuple, functor);
    }

    // Чтение PDO замапленных каждым слейвом
    template<typename... DataTypes>
    void read(DataTypes&... data) {
        if(sizeof...(data) > pdo_fmmu_slaves_.size()) {
            return;
        }

        size_t i = 0;
        custom_tuple<datagram::EcatDatagram<command::LRD, DataTypes>...> datagram_tuple{pdo_fmmu_slaves_[i++].make_datagram_read(data)...};
        DatagramFunctor functor{ecat_buffer_.get_ecat().get_telegram(), pdo_fmmu_slaves_.size(), nullptr};

        for_each_reverse(datagram_tuple, functor);
    }

    // Последовательность чтения-записи PDO замапленных каждым слейвом
    template<typename... DataTypes>
    void read_write(DataTypes&... data) {
        if(sizeof...(data) > pdo_fmmu_slaves_.size()) {
            return;
        }

        size_t i = 0;
        size_t j = 0;

        // custom_tuple с датаграммами и на чтение и на запись в две строки. 1 строка - тип, 2 строка - имя переменной
        custom_tuple<datagram::EcatDatagram<command::LRD, DataTypes>... , datagram::EcatDatagram<command::LWR, DataTypes>...>
        datagram_tuple{pdo_fmmu_slaves_[i++].make_datagram_read(data)... , pdo_fmmu_slaves_[j++].make_datagram_write(data)...};

        DatagramFunctor functor{ecat_buffer_.get_ecat().get_telegram(), pdo_fmmu_slaves_.size(), nullptr};

        for_each_reverse(datagram_tuple, functor);
    }
*/

};

} // namespace ecat_pdo_fmmu

} // namespace free_rtos

#endif /* FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_PDO_FMMU_HPP_ */
