/*
 * eth_ecat_datagram.hpp
 *
 *  Created on: May 2, 2023
 *      Author: algin
 */

#ifndef FREE_RTOS_ETHERNET_INDUSTRY_ETH_ECAT_DATAGRAM_HPP_
#define FREE_RTOS_ETHERNET_INDUSTRY_ETH_ECAT_DATAGRAM_HPP_

#include <cstdint>

#include "ethernet_industry/ethercattype.hpp"
#include "ethernet_industry/eth_ecat_types.h"
#include "ethernet_industry/eth_ecat_queue.hpp"
#include "ethernet_industry/eth_ecat_custom_tuple.hpp"
#include "ethernet_industry/eth_ecat_packer.hpp"
#include "ethernet_industry/eth_ecat_command.hpp"

namespace free_rtos {

namespace datagram {

using TEcatWkc = uint16_t;

class IEcatDatagram {
public:
    IEcatDatagram(ec_moredatagrams more, TEcatWkc wkc)
    : more_{more}
    , wkc_{wkc}
    { }

    IEcatDatagram() { }

    virtual ~IEcatDatagram() { };

    IEcatDatagram* get_next() {
        queue::QueueEntity<IEcatDatagram>* next = queue_entity_.get_next();

        if(next == nullptr) {
            return nullptr;
        }

        return next->get_data();
    }

    queue::QueueEntity<IEcatDatagram>& get_queue_entity() {
        return queue_entity_;
    }

    IEcatDatagram& operator+(IEcatDatagram &next) {
        more_ = ec_moredatagrams::EC_MOREDATAGRAMS_MORE;

        queue_entity_ + next.get_queue_entity();

        return next;
    }

    virtual uint8_t* pack(uint8_t *raw) = 0;
    virtual uint8_t* unpack(uint8_t *raw) = 0;

    TEcatDgHeader& get_header() {
        return header_;
    }

    TEcatWkc get_wkc() {
        return wkc_;
    }

    TEcatWkc get_all_wkc() {
        queue::QueueEntity<IEcatDatagram>* next = queue_entity_.get_next();

        if(next == nullptr) {
            return wkc_;
        }

        return wkc_ + next->get_data()->get_all_wkc();
    }

protected:
    ec_moredatagrams more_;
    TEcatDgHeader header_;
    TEcatWkc wkc_;

private:
    queue::QueueEntity<IEcatDatagram> queue_entity_{this};
};

template<typename CommandT, typename... DataTypes>
class EcatDatagram : public IEcatDatagram {
    static_assert(std::is_base_of<command::EcatCommandBase, CommandT>::value == true, "CommandT should be derived from ECatCommandBase");

public:
    EcatDatagram(CommandT&& command, DataTypes&... data)
    : IEcatDatagram{ec_moredatagrams::EC_MOREDATAGRAMS_LAST, 0x0000}
    , command_{command}
    , data_tuple_{data...} { }

    EcatDatagram() { }

    ~EcatDatagram() { }

    virtual uint8_t* pack(uint8_t *raw) override {
        return pack_header(raw);
    }

    virtual uint8_t* unpack(uint8_t *raw) override {
        return unpack_header(raw);
    }

private:
    CommandT command_;
    custom_tuple<DataTypes&...> data_tuple_;

    uint8_t* pack_wkc(uint8_t *raw) {
        TEcatWkc *wkc = new(raw) TEcatWkc{0x0000};

        (void)wkc;

        return raw + sizeof(TEcatWkc);
    }

    uint8_t* pack_data(uint8_t *raw) {
        PackFunctor functor{raw};

        for_each(data_tuple_, functor);

        return functor.raw;
    }

    uint8_t* pack_header(uint8_t *raw) {
        uint8_t *data_raw = raw + sizeof(TEcatDgHeader);
        uint8_t *wkc_raw = pack_data(data_raw); // сначала упаковываем  все данные для вычисления их размера
        uint16_t len = wkc_raw - data_raw;      // вычисляем размер данных
        TEcatDgHeader *header_ = new(raw) TEcatDgHeader{
            command_.get_cmd(),
            0x00,
            command_.get_address(),
            {.bits = {
                      .len = len,
                      .c = static_cast<uint16_t>(ec_circframe::EC_CIRCFRAME_NOT),
                      .m = static_cast<uint16_t>(more_)}},
            0x0000};

        (void)header_;

        return pack_wkc(wkc_raw);
    }

    uint8_t* unpack_wkc(uint8_t *raw) {
        TEcatWkc *wkc = reinterpret_cast<TEcatWkc*>(raw);

        wkc_ = *wkc;

        return raw + sizeof(TEcatWkc);
    }

    uint8_t* unpack_data(uint8_t *raw) {
        UnpackFunctor functor{raw};

        for_each(data_tuple_, functor);

        return functor.raw;
    }

    uint8_t* unpack_header(uint8_t *raw) {
        TEcatDgHeader *header = reinterpret_cast<TEcatDgHeader*>(raw);

        header_ = *header;

        return unpack_data(raw + sizeof(TEcatDgHeader));
    }
};

}


}



#endif /* FREE_RTOS_ETHERNET_INDUSTRY_ETH_ECAT_DATAGRAM_HPP_ */
